@@ -1,45 +1,54 @@
 """
 See docstring class Validator below for more details on validation
 """
+
 from abc import abstractmethod
 from copy import deepcopy
 from typing import Any, Callable, Dict, List, Type, Union
 
 from moto.dynamodb.exceptions import (
-    AttributeIsReservedKeyword,
-    ExpressionAttributeValueNotDefined,
     AttributeDoesNotExist,
-    ExpressionAttributeNameNotDefined,
-    IncorrectOperandType,
-    InvalidUpdateExpressionInvalidDocumentPath,
-    ProvidedKeyDoesNotExist,
+    AttributeIsReservedKeyword,
+    DuplicateUpdateExpression,
     EmptyKeyAttributeException,
-    UpdateHashRangeKeyException,
+    ExpressionAttributeNameNotDefined,
+    ExpressionAttributeValueNotDefined,
+    IncorrectOperandType,
+    InvalidAttributeTypeError,
+    InvalidUpdateExpressionInvalidDocumentPath,
     MockValidationException,
+    ProvidedKeyDoesNotExist,
+    TooManyClauses,
+    UpdateHashRangeKeyException,
 )
 from moto.dynamodb.models.dynamo_type import DynamoType, Item
 from moto.dynamodb.models.table import Table
 from moto.dynamodb.parsing.ast_nodes import (  # type: ignore
-    Node,
-    ExpressionAttribute,
-    UpdateExpressionClause,
-    UpdateExpressionPath,
-    UpdateExpressionSetAction,
-    UpdateExpressionAddAction,
-    UpdateExpressionDeleteAction,
-    UpdateExpressionRemoveAction,
     DDBTypedValue,
-    ExpressionAttributeValue,
-    ExpressionAttributeName,
     DepthFirstTraverser,
-    NoneExistingPath,
-    UpdateExpressionFunction,
+    ExpressionAttribute,
+    ExpressionAttributeName,
+    ExpressionAttributeValue,
     ExpressionPathDescender,
-    UpdateExpressionValue,
-    ExpressionValueOperator,
     ExpressionSelector,
+    ExpressionValueOperator,
+    Node,
+    NoneExistingPath,
+    UpdateExpression,
+    UpdateExpressionAddAction,
+    UpdateExpressionAddClause,
+    UpdateExpressionClause,
+    UpdateExpressionDeleteAction,
+    UpdateExpressionFunction,
+    UpdateExpressionPath,
+    UpdateExpressionRemoveAction,
+    UpdateExpressionRemoveClause,
+    UpdateExpressionSetAction,
+    UpdateExpressionSetClause,
+    UpdateExpressionValue,
 )
 from moto.dynamodb.parsing.reserved_keywords import ReservedKeywords
+from moto.dynamodb.utils import find_duplicates, find_path_overlaps
 
 
 class ExpressionAttributeValueProcessor(DepthFirstTraverser):  # type: ignore[misc]
@@ -255,7 +264,9 @@ class UpdateExpressionFunctionEvaluator(DepthFirstTraverser):  # type: ignore[mi
             raise NotImplementedError(f"Unsupported function for moto {function_name}")
 
     @classmethod
-    def get_list_from_ddb_typed_value(cls, node: DDBTypedValue, function_name: str) -> DynamoType:  # type: ignore[misc]
+    def get_list_from_ddb_typed_value(  # type: ignore[misc]
+        cls, node: DDBTypedValue, function_name: str
+    ) -> DynamoType:
         assert isinstance(node, DDBTypedValue)
         dynamo_value = node.get_value()
         assert isinstance(dynamo_value, DynamoType)
@@ -320,7 +331,9 @@ class ExecuteOperations(DepthFirstTraverser):  # type: ignore[misc]
         return dynamo_value
 
     @classmethod
-    def get_sum(cls, left_operand: DynamoType, right_operand: DynamoType) -> DDBTypedValue:  # type: ignore[misc]
+    def get_sum(  # type: ignore[misc]
+        cls, left_operand: DynamoType, right_operand: DynamoType
+    ) -> DDBTypedValue:
         """
         Args:
             left_operand(DynamoType):
@@ -335,7 +348,9 @@ class ExecuteOperations(DepthFirstTraverser):  # type: ignore[misc]
             raise IncorrectOperandType("+", left_operand.type)
 
     @classmethod
-    def get_subtraction(cls, left_operand: DynamoType, right_operand: DynamoType) -> DDBTypedValue:  # type: ignore[misc]
+    def get_subtraction(  # type: ignore[misc]
+        cls, left_operand: DynamoType, right_operand: DynamoType
+    ) -> DDBTypedValue:
         """
         Args:
             left_operand(DynamoType):
@@ -379,6 +394,34 @@ class EmptyStringKeyValueValidator(DepthFirstTraverser):  # type: ignore[misc]
         return node
 
 
+class TypeMismatchValidator(DepthFirstTraverser):  # type: ignore[misc]
+    def __init__(self, key_attributes_type: List[Dict[str, str]]):
+        self.key_attributes_type = key_attributes_type
+
+    def _processing_map(
+        self,
+    ) -> Dict[
+        Type[UpdateExpressionSetAction],
+        Callable[[UpdateExpressionSetAction], UpdateExpressionSetAction],
+    ]:
+        return {UpdateExpressionSetAction: self.check_for_type_mismatch}
+
+    def check_for_type_mismatch(
+        self, node: UpdateExpressionSetAction
+    ) -> UpdateExpressionSetAction:
+        """A node representing a SET action. Check that type matches with the definition"""
+        assert isinstance(node, UpdateExpressionSetAction)
+        assert len(node.children) == 2
+        key = node.children[0].children[0].children[0]
+        val_node = node.children[1].children[0]
+        for dct in self.key_attributes_type:
+            if dct["AttributeName"] == key and dct["AttributeType"] != val_node.type:
+                raise InvalidAttributeTypeError(
+                    key, dct["AttributeType"], val_node.type
+                )
+        return node
+
+
 class UpdateHashRangeKeyValidator(DepthFirstTraverser):  # type: ignore[misc]
     def __init__(
         self,
@@ -409,6 +452,82 @@ class UpdateHashRangeKeyValidator(DepthFirstTraverser):  # type: ignore[misc]
         return node
 
 
+class OverlappingPathsValidator(DepthFirstTraverser):  # type: ignore[misc]
+    def __init__(self, expression_attribute_names: Dict[str, str]):
+        self.expression_attribute_names = expression_attribute_names
+
+    def _processing_map(
+        self,
+    ) -> Dict[
+        Type[UpdateExpression],
+        Callable[[UpdateExpression], UpdateExpression],
+    ]:
+        return {UpdateExpression: self.check_for_overlapping_paths}
+
+    def check_for_overlapping_paths(self, node: UpdateExpression) -> UpdateExpression:
+        actions = node.find_clauses(
+            [
+                UpdateExpressionSetAction,
+                UpdateExpressionAddAction,
+                UpdateExpressionRemoveAction,
+                UpdateExpressionDeleteAction,
+            ]
+        )
+        expression_attribute_name_paths = []
+        plaintext_paths = []
+        for action in actions:
+            path = action.children[0]
+            path_name = path.children[0]
+            # for some reason, if we only have a root expression ('#path1', '#path2'),
+            # AWS only checks for direct matches, not path overlaps
+            if (
+                isinstance(path_name, ExpressionAttributeName)
+                and len(path.children) == 1
+            ):
+                expression_attribute_name_paths.append(
+                    self.expression_attribute_names[
+                        path_name.get_attribute_name_placeholder()
+                    ]
+                )
+            else:
+                plaintext_paths.append(path.to_str())
+
+        overlapping_paths = find_path_overlaps(plaintext_paths)
+        if overlapping_paths:
+            # There might be more than one attribute duplicated:
+            # they may get mixed up in the Error Message which is inline with actual boto3 Error Messages
+            raise DuplicateUpdateExpression(*overlapping_paths)
+
+        duplicates = find_duplicates(plaintext_paths + expression_attribute_name_paths)
+        if duplicates:
+            raise DuplicateUpdateExpression(*duplicates)
+        return node
+
+
+class ActionCountValidator(DepthFirstTraverser):  # type: ignore[misc]
+    def _processing_map(
+        self,
+    ) -> Dict[
+        Type[UpdateExpression],
+        Callable[[UpdateExpression], UpdateExpression],
+    ]:
+        return {UpdateExpression: self.verify_action_counts}
+
+    def verify_action_counts(self, node: UpdateExpression) -> UpdateExpression:
+        add_clauses = node.find_clauses([UpdateExpressionAddClause])
+        remove_clauses = node.find_clauses([UpdateExpressionRemoveClause])
+        set_clauses = node.find_clauses([UpdateExpressionSetClause])
+
+        # Only allow single ADD/REMOVE clauses
+        if len(add_clauses) > 1:
+            raise TooManyClauses("ADD")
+        if len(remove_clauses) > 1:
+            raise TooManyClauses("REMOVE")
+        if len(set_clauses) > 1:
+            raise TooManyClauses("SET")
+        return node
+
+
 class Validator:
     """
     A validator is used to validate expressions which are passed in as an AST.
@@ -454,6 +573,8 @@ class UpdateExpressionValidator(Validator):
     def get_ast_processors(self) -> List[DepthFirstTraverser]:
         """Get the different processors that go through the AST tree and processes the nodes."""
         processors = [
+            ActionCountValidator(),
+            OverlappingPathsValidator(self.expression_attribute_names),
             UpdateHashRangeKeyValidator(
                 self.table.table_key_attrs, self.expression_attribute_names or {}
             ),
@@ -464,6 +585,7 @@ class UpdateExpressionValidator(Validator):
             UpdateExpressionFunctionEvaluator(),
             NoneExistingPathChecker(),
             ExecuteOperations(),
+            TypeMismatchValidator(self.table.attr),
             EmptyStringKeyValueValidator(self.table.attribute_keys),
         ]
         return processors
